<?php

namespace App\Http\Controllers\API;

use App\Http\Controllers\API\BaseController as BaseController;
use Illuminate\Support\Facades\Validator;
use Illuminate\Http\Request;
use Illuminate\Validation\Rule;
use App\Models\Agent;
use App\Models\Company;
use App\Helpers\Utilities;
use App\Helpers\MoviderHelper;

class AgentController extends BaseController
{

    private function privateReader($companyid=0, $id=0, $user=[])
    {
        try {
            $company = Company::where('id', $companyid)->where('user_id', $user->id)->first();
            if ($company) {
                $agent = Agent::where('id', $id)->where('company_id', $company->id)->with('company')->first();
                if ($agent) {
                    return ['status' => 'success', 'agent' => $agent];
                } else {
                    return ['status' => 'error', 'type' => 'availability'];
                }
            } else {
                return ['status' => 'error', 'type' => 'non-access'];
            }

        } catch (\Exception $e) {
            \Log::error($e);
        }
    }
    public function readAgent(Request $request, $companyid=0, $id=0)
    {
        try {
            $data = $request->all();
            $user = $request->user();
            $response = $this->privateReader($companyid, $id,$user);
            if ($response['status'] == 'success') {
                return $this->sendResponse(['result' => $response['agent']], 'Here is the agent details.');
            } else {
                if ($response['type'] == 'availability') {
                    $this->sendError("This agent does not exist.", [], 406);
                } else {
                    $this->sendError("You do not have access to this agent.", [], 406);
                }
            }
        } catch (\Exception $e) {
            \Log::error($e);
        }
    }

    public function create(Request $request)
    {
        try {
            $data = $request->all();
            $user = $request->user();
            $validator = Validator::make($data, [
                'mode' => 'required|in:validate,create',
                'name' => 'required',
                'number' => ['required', 'numeric', Rule::unique('agents')->whereNull('deleted_at')],
                'company_id' => 'required|numeric|exists:company,id']);
            if ($validator->fails()) {
                return $this->sendError($validator->errors()->first(), [], 406);
            }

            if ($data['mode'] == 'validate') {
                $result = (new MoviderHelper)->verifyAccount(trim($data['number']));
                if ($result['error'] && !is_null($result['error'])) {
                    \Log::error($result['error']);
                    return $this->sendError('Check your phone number format. It should be a valid phone number.', $result['error'], 406);
                } else {
                    $responseJson = json_decode($result['response']);
                    if (isset($responseJson->error) && isset($responseJson->error->code) && isset($responseJson->error->description)) {
                        return $this->sendError($responseJson->error->description, [], 406);
                    } else {
                        return $this->sendResponse(json_decode($result['response']), "We sent an OTP to your provided phone number. Please check to complete this process.");
                    }
                }
            }

            $extra = date('Y-m-d H:i:s').$user->id.$data['company_id'];
            $newagent = new Agent();
            $newagent->company_id = $data['company_id'];
            $newagent->unique_key = (new Utilities)->generateHash($extra);
            $newagent->name = $data['name'];
            $newagent->number = $data['number'];
            $newagent->status = 'offline';
            $newagent->last_available_at = null;
            $newagent->save();

            $agent = Agent::where('id', $newagent->id)->with('company')->first();

            return $this->sendResponse(['result' => $agent], 'Here is the new agent details.');

        } catch (\Exception $e) {
            \Log::error($e);
        }
    }

    public function edit(Request $request)
    {
        try {
            $data = $request->all();
            $user = $request->user();
            $rules = [
                'company_id' => 'required|exists:company,id',
                'id' => 'required|exists:agents,id'
            ];
            if (isset($data['name'])) {
                $rules['name'] = 'required';
            }
            if (isset($data['number'])) {
                $rules['number'] = 'required|numeric';
            }

            if (isset($data['agent_status'])) {
                $rules['agent_status'] = 'required|in:online,offline';
            }
            $validator = Validator::make($data, $rules);
            if ($validator->fails()) {
                return $this->sendError($validator->errors()->first(), [], 406);
            }

            $response = $this->privateReader($data['company_id'], $data['id'], $user);
            if ($response['status'] == 'success') {
                if (isset($data['name'])) {
                    $response['agent']->name = $data['name'];
                }
                if (isset($data['number'])) {
                    $response['agent']->number = $data['number'];
                }

                if (isset($data['agent_status']) && !empty($data['agent_status'])) {
                    $response['agent']->status = $data['agent_status'];
                    $response['agent']->last_available_at = date('Y-m-d H:i:s');
                }
                $response['agent']->save();
                return $this->sendResponse(['result' => $response['agent']], 'Agent details has been updated.');
            } else {
                if ($response['type'] == 'availability') {
                    $this->sendError("This agent does not exist.", [], 406);
                } else {
                    $this->sendError("You do not have access to this agent.", [], 406);
                }
            }

        } catch (\Exception $e) {
            \Log::error($e);
        }
    }

    public function delete(Request $request)
    {
        try {
            $data = $request->all();
            $user = $request->user();
            $validator = Validator::make($data, [
                'company_id' => 'required|numeric|exists:company,id',
                'id' => 'required|numeric|exists:agents,id'
            ]);
            if ($validator->fails()) {
                return $this->sendError($validator->errors()->first(), [], 406);
            }

            $response = $this->privateReader($data['company_id'], $data['id'], $user);
            if ($response['status'] == 'success') {
                $response['agent']->forceDelete();
                return $this->sendResponse([], 'Agent has been deleted');
            } else {
                if ($response['type'] == 'availability') {
                    $this->sendError("This agent does not exist.", [], 406);
                } else {
                    $this->sendError("You do not have access to this agent.", [], 406);
                }
            }

        } catch (\Exception $e) {
            \Log::error($e);
        }
    }
}
