<?php

namespace App\Http\Controllers\API;

use App\Http\Controllers\API\BaseController as BaseController;
use Illuminate\Support\Facades\Validator;
use Illuminate\Http\Request;
use App\Models\Agent;
use App\Models\Call;
use App\Models\Company;
use App\Helpers\Utilities;
use App\Events\PhoneCallAvailable;

class CallController extends BaseController
{
    private function privateReader($ds=[])
    {
        try {
            $company = Company::where('unique_key', $ds['company_unique_key'])->where('user_id', $ds['user']->id)->first();
            if ($company) {
                $call = Call::where('id', $ds['id'])->where('company_unique_key', $ds['company_unique_key'])->first();
                if ($call) {
                    return ['status' => 'success', 'call' => $call];
                } else {
                    return ['status' => 'error', 'type' => 'availability'];
                }
            } else {
                return ['status' => 'error', 'type' => 'non-access'];
            }

        } catch (\Exception $e) {
            \Log::error($e);
        }
    }

    public function readCall(Request $request, $company_unique_key='', $id=0)
    {
        try {
            $data = $request->all();
            $user = $request->user();
            $response = $this->privateReader([
                'company_unique_key' => $company_unique_key,
                'id' => $id,
                'user' => $user
            ] );
            if ($response['status'] == 'success') {
                return $this->sendResponse(['result' => $response['call']], 'Here is the call details.');
            } else {
                if ($response['type'] == 'availability') {
                    $this->sendError("This call does not exist.", [], 406);
                } else {
                    $this->sendError("You do not have access to this call.", [], 406);
                }
            }

        } catch (\Exception $e) {
            \Log::error($e);
        }
    }

    public function readList(Request $request, $company_unique_key='')
    {
        try {
            $data = $request->all();
            $user = $request->user();
            $company = Company::where('unique_key', $company_unique_key)->where('user_id', $user->id)->first();
            if ($company) {
                $calls = Call::where('company_unique_key', $company_unique_key)->get();
                return $this->sendResponse(['result' => $calls], 'Here is the call list details.');
            } else {
                $this->sendError("You do not have access to this list.", [], 406);
            }
        } catch (\Exception $e) {
            \Log::error($e);
        }
    }

    public function create(Request $request)
    {
        try {
            $data = $request->all();
            $user = $request->user();
            $validator = Validator::make($data, [
                'agent_id' => 'required|exists:agents,id',
                'name' => 'required',
                'number' => 'required|numeric',
                'company_unique_key' => 'required'
            ]);

            if ($validator->fails()) {
                return $this->sendError('Check these fields first before completing this request.', $validator->errors(), 406);
            }
            $data['user'] = $user;
            return $this->processSave($data);

        } catch (\Exception $e) {
            \Log::error($e);
        }
    }

    private function processSave($data=[], $mode='create')
    {
        try {
            $company = Company::where('unique_key', $data['company_unique_key'])->where('user_id', $data['user']->id)->first();
            if ($company) {
                $agent = Agent::where('id', $data['agent_id'])->where('company_id', $company->id)->first();
                if ($agent) {
                    if ($mode == 'edit') {
                        $call = Call::where('id', $data['call_id'])->first();
                        if (isset($data['call_id']) && !empty($data['call_id']) && !empty($call)) {
                            // \Log::info($data);
                            $haschanges = false;
                            if (isset($data['status'])) {
                                // \Log::info($call);
                                // \Log::info($data);
                                $call->status = $data['status'];
                                $haschanges = true;
                            }

                            if (isset($data['media_id'])) {
                                $call->media_id = $data['media_id'];
                                $haschanges = true;
                            }

                            if (isset($data['name'])) {
                                $call->call_name = $data['name'];
                                $haschanges = true;
                            }

                            if (isset($data['number'])) {
                                $call->call_number = $data['number'];
                                $haschanges = true;
                            }

                            if ($haschanges) {
                                $call->save();
                            }
                            return $this->sendResponse(['result' => $call], 'Call details has been updated.');
                        } else {
                            \Log::info($call);
                            return $this->sendError("This call does not exist.", [], 406);
                        }
                    } else {
                        $call = Call::where('call_number', $data['number'])->where('company_unique_key', $data['company_unique_key'])->first();
                        if ($call) {
                            return $this->sendError("This number already exist within your company. Try another number for this call data.", [], 406);
                        } else {
                            $newcall = new Call();
                            $newcall->company_id = $company->id;
                            $newcall->company_unique_key = $company->unique_key;
                            $newcall->agent_id = $agent->id;
                            $newcall->status = 'pending';
                            $newcall->media_id = null;
                            $newcall->call_name = $data['name'];
                            $newcall->call_number = $data['number'];
                            $newcall->save();
                            return $this->sendResponse(['result' => $newcall], 'Call details has been created.');
                        }
                    }
                } else {
                    return $this->sendError("You are not allowed to create call data. The agent does not exist anymore.", [], 406);
                }
            } else {
                return $this->sendError("You are not allowed to create call data based on your company key.", [], 406);
            }

        } catch (\Exception $e) {
            \Log::error($e);
        }
    }

    public function edit(Request $request)
    {
        try {
            $data = $request->all();
            $user = $request->user();
            $validFields = [
                'call_id' => 'required|exists:calls,id',
                'agent_id' => 'required|exists:agents,id',
                'company_unique_key' => 'required'
            ];

            if (isset($data['name'])) {
                $validFields['name'] = 'required';
            }

            if (isset($data['number'])) {
                $validFields['name'] = 'required|required';
            }

            $validator = Validator::make($data, $validFields);
            if ($validator->fails()) {
                return $this->sendError('Check these fields first before completing this request.', $validator->errors(), 406);
            }
            $data['user'] = $user;
            return $this->processSave($data, 'edit');

        } catch (\Exception $e) {
            \Log::error($e);
        }
    }

    public function delete(Request $request)
    {
        try {
            $data = $request->all();
            $user = $request->user();
            $validator = Validator::make($data, [
                'agent_id' => 'required|exists:agent,id',
                'call_id' => 'required|exists:call,id',
                'company_unique_key' => 'required|exists:call,company_unique_key|exists:company,company_unique_key'
            ]);
            if ($validator->fails()) {
                return $this->sendError('Check these fields first before completing this request.', $validator->errors(), 406);
            }

            $response = $this->privateReader([
                'company_unique_key' => $data['company_unique_key'],
                'id' => $data['call_id'],
                'user' => $user
            ]);
            if ($response['status'] == 'success') {
                $response['call']->delete();
                return $this->sendResponse([], 'Call has been deleted');
            } else {
                if ($response['type'] == 'availability') {
                    $this->sendError("This call does not exist.", [], 406);
                } else {
                    $this->sendError("You do not have access to this call.", [], 406);
                }
            }

        } catch (\Exception $e) {
            \Log::error($e);
        }
    }

    public function broadcastCall(Request $request)
    {
        try {
            $agent = Agent::where('id', 2)->first()->toArray();
            \Log::info('agent', $agent);
            $recipient = ['number' => '639175940318', 'name' => 'Raciel'];
            \Log::info('recipient', $recipient);
            $rawdata = ['agent'=> $agent, 'recipient' => $recipient];
            $data= ['agent'=> $agent, 'recipient' => $recipient, 'message'=>json_encode($rawdata)];
            \Log::info('WS DATA:', $data);
            \App\Events\Test::dispatch('This is a test message');
            event(new \App\Events\Test('This is a test message from events'));
        } catch (\Exception $e) {
            \Log::error($e);
        }
    }

    public function setCall(Request $request)
    {
        try {
            $data = $request->all();
            $user = $request->user();
            //Start with mock set
            $agent = Agent::where('id', 1)->with('company')->first();
            \Log::info('Agent', (array) $agent);
            $req = [
                'agent_id' => $agent->id,
                'name' => $agent->name,
                'number' => '09278693921',
                'company_unique_key' => $agent->company->unique_key
            ];
            \Log::info('Data', $req);
            $req['user'] = $user;
            return $this->processSave($req);

        } catch (\Exception $e) {
            \Log::error($e);
        }
    }

    public function agentCalls(Request $request, $id = 0)
    {
        try {
            $data = $request->all();
            $user = $request->user();
            if ($id > 0) {
                $calls = Call::where('agent_id', $id)->where('status', 'pending')->get();
                return $this->sendResponse($calls, 'Here are the agent set of calls');
            } else {
                $this->sendError("Agent does not have any call set.", [], 406);
            }
        } catch (\Exception $e) {
            \Log::error($e);
        }
    }
}
