<?php

namespace App\Http\Controllers\API;

use App\Http\Controllers\API\BaseController as BaseController;
use Illuminate\Support\Facades\Validator;
use Illuminate\Http\Request;
use App\Models\User;
use App\Helpers\UserHelper;
use App\Helpers\ProfileHelper;
use App\Helpers\SecurityHelper;
use Chikolokoy08\Dengine\Dengine;
use SimpleSoftwareIO\QrCode\Facades\QrCode;
class CashierController extends BaseController
{
    public function list(Request $request)
    {
        try {

            $data = $request->all();
            $parent_id = $request->user()->id;

            $dengine = new Dengine();
            $dengine::parse($data);

            //MODEL INSTANCE
            $model = new User();
            $dtparams = $dengine::setParameters(['parent_id' => $parent_id]);

            //TOTAL RECORDS QUERY
            $total = count($model->getCashiers($dtparams));

            //QUERY WITH FULL PARAMETERS
            $query = $model->getCashiers($dtparams);

            //SETTER FOR TOTAL
            $dengine::recordsTotal($total);

            //PREPARING QUERY RESULT INTO DATATABLE JSON FORMAT
            $dengine::prepare($query);

            //RETURN AS JSON OBJECT
            return $dengine::make();

        } catch (\Exception $e) {
            \Log::error($e);
        }
    }

    public function detail(Request $request, $id=0)
    {
        try {
            $user = $request->user();
            $cashier = User::where('id', $id)->first();
            if (is_null($cashier)) {
                return $this->sendError('Record not found.', [], 406);
            }
            return $this->sendResponse($cashier, 'Cashier data.');

        } catch (\Exception $e) {
            \Log::error($e);
        }
    }

    public function create(Request $request)
    {
        try {
            $data = $request->all();
            $user = $request->user();
            $data['parent_id'] = $user->id;
            // \Log::info('USER ON CASHIER CREATE', (array) $user);
            $validator = Validator::make($data, [
                'firstname' => 'required',
                'lastname' => 'required',
                'email' => 'email:rfc,dns|unique:users,email',
                'password'=>'required|confirmed',
                'password_confirmation'=>'required',
                'account_validation' => 'required',
                'newsletter_validation' => 'required'
            ]);

            if ($validator->fails()) {
                return $this->sendError('Check these fields first before completing this request.', $validator->errors()->first(), 406);
            }

            $new_user = UserHelper::createCashier($data, $user->id);
            $new_user->status = true;
            $new_user->email_verified_at = date('Y-m-d H:i:s');
            $new_user->verified_at = date('Y-m-d H:i:s');
            $new_user->save();
            $profile = ProfileHelper::createProfile($new_user, $data);

            if ($new_user) {
                if ($data['account_validation'] == 'yes') {
                    //Send email for retrieval of password
                    $passwordReset = SecurityHelper::requestPasswordReset($data['email']);
                    \Mail::to($data['email'])->send(new \App\Mail\PasswordResetRequest([
                        'title' => 'Reset Password to confirm your new Cashier account',
                        'appname' => env('APP_NAME'),
                        'email' => $data['email'],
                        'token' => $passwordReset->token
                    ]));
                    return $this->sendResponse(['profile' => $profile], 'One last thing. We sent a token to your email. Use that to verify the account you created to complete the sign up process. Thanks.');
                } else {
                    //Send email but verified already
                    $details = [
                        'title' => 'Welcome to '.env('APP_NAME'),
                        'appname' => env('APP_NAME'),
                        'firstname' => $data['firstname'],
                        'email' => $data['email'],
                        'code' => $profile->signup_code,
                        'serialize' => $profile->signup_serialize,
                        'signup_step' => 'two',
                        'qrcode' => UserHelper::generateQrCode($profile->qrstring)
                    ];

                    \Mail::to($data['email'])->send(new \App\Mail\SignupEmail($details));
                }
                return $this->sendResponse(['profile' => $profile], 'Cashier account is created. Thanks.');

            } else {
                return $this->sendError('Cashier creation failed. Check admin for support.', [], 406);
            }


        } catch (\Exception $e) {
            \Log::error($e);
        }
    }

    public function update(Request $request)
    {
        try {
            $data = $request->all();
            $user = $request->user();
            $rules = [
                'email' => 'required|exists:users,email',
                'firstname' => 'required',
                'lastname' => 'required',
                'newsletter_validation' => 'required'
            ];
            if (isset($data['password']) && !empty($data['password'])) {
                $rules['password'] = 'required|confirmed';
                $rules['password_confirmation'] = 'required';
            }

            $validator = Validator::make($data, $rules);

            if ($validator->fails()) {
                return $this->sendError('Check these fields first before completing this request.', $validator->errors()->first(), 406);
            }

            //Get user object
            $user = User::where('email', $data['email'])->with('profile')->first();

            if (is_null($user)) {
                return $this->sendError('Check these fields first before completing this request.', 'User not found', 406);
            }

            if (isset($data['password']) && !empty($data['password'])) {
                $user->password = $data['password'];
                $user->save();
            }

            if (isset($data['mobile']) && !empty($data['mobile'])) {
                $user->profile->mobile = $data['mobile'];
            }

            if (isset($data['address']) && !empty($data['address'])) {
                $user->profile->address = $data['address'];
            }
            $user->profile->save();
            return $this->sendResponse(['profile' => $user->profile], 'Cashier account has been updated. Thanks.');

        } catch (\Exception $e) {
            \Log::error($e);
        }
    }
    public function delete(Request $request)
    {
        try {
            $data = $request->all();
            $user = $request->user();
            $validator = Validator::make($data, ['user_id' => 'required|exists:users,id']);

            if ($validator->fails()) {
                return $this->sendError('Check these fields first before completing this request.', $validator->errors()->first(), 406);
            }

            $userdata = UserHelper::deleteUser($data, 'cashier');
            \Log::info('USER DELETION', $userdata);
            if ($userdata['result']) {
                return $this->sendResponse([], 'Cashier data is successfully deleted.');
            } else {
                return $this->sendError('Cashier data not found.', [], 406);
            }

        } catch (\Exception $e) {
            \Log::error($e);
        }
    }
}
