<?php

namespace App\Http\Controllers\API;

use App\Helpers\SecurityHelper;
use App\Http\Controllers\API\BaseController as BaseController;
use App\Helpers\UserHelper;
use App\Helpers\ProfileHelper;
use App\Helpers\Utilities;
use App\Models\UserProfile;
use App\Models\PasswordReset;
use Illuminate\Http\Request;
use App\Models\Invite;
use Illuminate\Support\Facades\Validator;


use App\Models\User;
use App\Models\Company;
use Illuminate\Support\Facades\Hash;
use Illuminate\Support\Str;

use SimpleSoftwareIO\QrCode\Facades\QrCode;

class SecurityController extends BaseController
{
    public function loginUser(Request $request)
    {
        try {
            $data = $request->all();
            // \Log::info($data);
            $validator = Validator::make($data, ['email' => 'email:rfc,dns','password'=>'required']);
            if ($validator->errors()->first()) {
                return $this->sendError('Check these fields first before completing the login request.', $validator->errors()->first(), 406);
            }

            if(\Auth::attempt(['email' => $data['email'], 'password' => $data['password']])){
                $user = \Auth::user();
                $success['token'] =  $user->createToken($user->email)->plainTextToken;
                $success['user'] = $user;
                $success['redirect'] = 'dashboard';
                $user->last_login = date('Y-m-d H:i:s');
                $user->save();
                // $success['company'] = Company::where('user_id', $user->id)->get();
                return $this->sendResponse($success, 'Login is done.');
            }

            return $this->sendError('Login failed. Email / Password is invalid', [], 406);


        } catch (\Exception $e) {
            \Log::error($e);
        }
    }

    public function registerUser(Request $request) {
        try {
            //Check if invite code is valid and hasn't been activated
            $data = $request->all();
            // $valid_code = Invite::where('code', $data['code'])->first();
            // if (!empty($code) && $valid_code && $valid_code->status) {
            //     return $this->sendError('Invite code has been activate already. You can not create a user with activated code.', [], 406);
            // }
            //Check if email and password is empty
            $validator = Validator::make($data, ['email' => 'email:rfc,dns|unique:users','password'=>'required|confirmed', 'password_confirmation'=>'required']);
            if ($validator->errors()->first()) {
                return $this->sendError('Check these fields first before completing Sign up process.', $validator->errors()->first(), 406);
            }

            //Create User return token
            $new_user = UserHelper::createUser($data);
            $profile = ProfileHelper::createProfile($new_user, $data);
            // $valid_code->status = 1;
            // $valid_code->save();

            if ($new_user) {
                $details = [
                    'title' => 'Welcome to '.env('APP_NAME'),
                    'appname' => env('APP_NAME'),
                    'firstname' => $data['firstname'],
                    'email' => $data['email'],
                    'code' => $profile->signup_code,
                    'serialize' => $profile->signup_serialize,
                    'signup_step' => 'one'
                ];
                \Mail::to($data['email'])->send(new \App\Mail\SignupEmail($details));

                return $this->sendResponse(['profile' => $profile], 'One last thing. We sent a code to your email. Use that to verify your account to complete the sign up process. Thanks.');
            } else {
                return $this->sendError('Signup failed. Check admin for support.', [], 406);
            }

        } catch (\Exception $e) {
            \Log::error($e);
        }
    }

    public function verifyCode(Request $request)
    {
        try {
            $data = $request->all();
            $validator = Validator::make($data, [
                'signup_code' => 'required',
                'signup_serialize' => 'required',
            ]);

            if ($validator->errors()->first()) {
                return $this->sendError('Invalid code please call admin to resolve this issue. Thank you.', $validator->errors()->first(), 406);
            }

            try {
                $objProfile = (new Utilities)->encdecData('de', $data['signup_serialize']);
                $profile = UserProfile::where('user_id', $objProfile->user_id)->where('signup_code', $data['signup_code'])->first();
                if (is_null($profile)) {
                    return $this->sendError('Invalid code please call admin to resolve this issue. Thank you.', [], 406);
                }
                // \Log::info($profile);
                //Update status to true
                $user = User::where('id', $profile->user_id)->first();
                $user->status = true;
                $user->email_verified_at = date('Y-m-d H:i:s');
                $user->verified_at = date('Y-m-d H:i:s');
                $user->save();

                $details = [
                    'title' => 'Welcome to '.env('APP_NAME'),
                    'appname' => env('APP_NAME'),
                    'firstname' => $profile->firstname,
                    'email' => $user->email,
                    'qrcode' => QrCode::size(200)->generate($profile->qrstring),
                    'signup_step' => 'two'
                ];
                //Send email
                \Mail::to($details['email'])->send(new \App\Mail\SignupEmail($details));
                return $this->sendResponse([], 'Congratulations! You can now login.');
            } catch (\Exception $e) {
                return $this->sendError('Invalid code please call admin to resolve this issue. Thank you.', $validator->errors()->first(), 406);
            }



        } catch (\Exception $e) {
            \Log::error($e);
        }
    }

    public function requestReset(Request $request)
    {
        try {
            $data = $request->all();
            $validator = Validator::make($data, [
                'email' => 'email:rfc,dns|exists:users,email'
            ]);

            if ($validator->fails()) {
                return $this->sendError('Check these fields first before completing this request.', $validator->errors()->first(), 406);
            }

            $passwordReset = SecurityHelper::requestPasswordReset($data['email']);
            $details = [
                'title' => env('APP_NAME').'| Reset Password',
                'appname' => env('APP_NAME'),
                'email' => $passwordReset->email,
                'token' => $passwordReset->token
            ];
            \Mail::to($data['email'])->send(new \App\Mail\PasswordResetRequest($details));
            return $this->sendResponse([], 'Password reset request successful.');

        } catch (\Exception $e) {
            \Log::error($e);
        }
    }

    public function verifyReset(Request $request)
    {
        try {
            $data = $request->all();
            $validator = Validator::make($data, [
                'token' => 'required|exists:password_resets,token',
                'email' => 'email:rfc,dns|exists:users,email'
            ]);
            if ($validator->fails()) {
                return $this->sendError('Check these fields first before completing this request.', $validator->errors()->first(), 406);
            }

            $isTokenValid = PasswordReset::where('token', $data['token'])->where('email', $data['email'])->first();
            if (is_null($isTokenValid)) {
                return $this->sendError('This request has expire.', [], 406);
            }

            $pr = PasswordReset::find($isTokenValid->id);
            $pr->delete();
            return $this->sendResponse([], 'Reset request is valid.');

        } catch (\Exception $e) {
            \Log::error($e);
        }
    }

    public function passwordReset(Request $request)
    {
        try {
            $data = $request->all();
            $validator = Validator::make($data, [
                'reset_id' => 'required|exists:password_resets,id',
                'email' => 'email:rfc,dns|exists:users,email',
                'password' => 'required|confirmed',
                'password_confirmation' => 'required'
            ]);
            if ($validator->fails()) {
                return $this->sendError('Check these fields first before completing this request.', $validator->errors()->first(), 406);
            }

            if ($data['password'] != $data['password_confirmation']) {
                return $this->sendError('Check these fields first before completing this request.', 'Passwords does not match.', 406);
            }

            $user = User::where('email', $data['email'])->first();
            if (is_null($user)) {
                return $this->sendError('Check these fields first before completing this request.', 'Request can not proceed. User not found.', 406);
            }
            $user->password = $data['password'];
            $user->save();
            $pr = PasswordReset::find($data['reset_id']);
            $pr->delete();
            return $this->sendResponse(['redirect' => 'dashboard'], 'Your password is now updated.');

        } catch (\Exception $e) {
            \Log::error($e);
        }
    }

    public function unauthorized(Request $request)
    {
        try {
            return $this->sendError('UNAUTHORIZED', ['error'=>'Unauthorised'], 403);
        } catch (\Exception $e) {
            \Log::error($e);
        }
    }

}
