<?php

namespace App\Http\Controllers\API;

use App\Http\Controllers\API\BaseController as BaseController;
use Illuminate\Support\Facades\Validator;
use Illuminate\Http\Request;
use App\Models\User;
use App\Models\UserProfile;
use App\Models\Transaction;
use App\Models\Service;
use App\Helpers\UserHelper;
use Chikolokoy08\Dengine\Dengine;
use App\Helpers\Utilities;

class TransactionController extends BaseController
{
    public function list(Request $request)
    {
        try {

            $data = $request->all();
            $user = $request->user();
            $user_id = $user->id;

            $dengine = new Dengine();
            $dengine::parse($data);

            //MODEL INSTANCE
            $model = new Transaction();
            $dtparams = $dengine::setParameters(['user_id' => $user_id, 'role' => $user->role]);
            //TOTAL RECORDS QUERY
            $total = count($model->getTransactionFeeds($dtparams));
            //QUERY WITH FULL PARAMETERS
            $query = $model->getTransactionFeeds($dtparams);
            //SETTER FOR TOTAL
            $dengine::recordsTotal($total);

            //PREPARING QUERY RESULT INTO DATATABLE JSON FORMAT
            $dengine::prepare($query);

            //RETURN AS JSON OBJECT
            return $dengine::make();

        } catch (\Exception $e) {
            \Log::error($e);
        }
    }

    public function detail(Request $request, $id=0)
    {
        try {
            $user = $request->user();
            $transaction = Transaction::where('id', $id)->with('user','cashier','profile','cashprof')->first();
            if (is_null($transaction)) {
                return $this->sendError('Record not found.', [], 406);
            }
            return $this->sendResponse($transaction, 'Transaction data.');

        } catch (\Exception $e) {
            \Log::error($e);
        }
    }

    public function create(Request $request)
    {
        try {
            $data = $request->all();
            $user = $request->user();
            $validator = Validator::make($data, [
                'services' => 'required|array|min:1',
                'customer_id' => 'required|exists:users,id',
            ]);

            if ($validator->fails()) {
                return $this->sendError('Check these fields first before completing this request.', $validator->errors()->first(), 406);
            }

            foreach($data['services'] as $sid) {
                $service = Service::where('id', $sid)->first();
                $transaction = new Transaction();
                $transaction->user_id = $data['customer_id'];
                $transaction->service_id = $service->id;
                $transaction->service = $service->name;
                $transaction->cashier_id = $user->id;
                $transaction->amount = $service->cost;
                $transaction->points = $service->points;
                $transaction->save();
            }
            $transactionTxt = count($data['services']) > 1 ? 'Transactions' : 'Transaction';
            return $this->sendResponse([], $transactionTxt.' data has been created.');


        } catch (\Exception $e) {
            \Log::error($e);
        }
    }

    public function update(Request $request)
    {
        try {
            $data = $request->all();
            $user = $request->user();
            $validator = Validator::make($data, [
                'service_id' => 'required|exists:services,id',
                'service' => 'required',
                'customer_id' => 'required|exists:users,id',
                'txn_id' => 'required|exists:transactions,id',
                'amount' => 'required|numeric|between:-9999999999.99,9999999999.99',
                'points' => 'required|digits_between:1,99999999999999',
            ]);

            if ($validator->fails()) {
                return $this->sendError('Check these fields first before completing this request.', $validator->errors()->first(), 406);
            }

            $transaction = Transaction::where('id', $data['txn_id'])->where('user_id', $data['customer_id'])->first();
            if (!is_null($transaction)) {
                $transaction->service = $data['service'];
                if (isset($data['points']) && !is_null($data['points'])) {
                    $transaction->points = $data['points'];
                }
                if (isset($data['amount']) && !is_null($data['amount'])) {
                    $transaction->amount = $data['amount'];
                }
                $transaction->save();
                return $this->sendResponse(['txn' => $transaction], 'Transaction data has been updated.');
            } else {
                return $this->sendError('Record not found.', [], 406);
            }
        } catch (\Exception $e) {
            \Log::error($e);
        }
    }

    public function delete(Request $request)
    {
        try {
            $data = $request->all();
            $user = $request->user();
            $validator = Validator::make($data, ['txn_id' => 'required|exists:transactions,id']);

            if ($validator->fails()) {
                return $this->sendError('Check these fields first before completing this request.', $validator->errors()->first(), 406);
            }

            $transaction = Transaction::where('id', $data['txn_id'])->first();
            if (!is_null($transaction)) {
                $transaction->delete();
                return $this->sendResponse([], 'Transaction data is successfully deleted.');
            } else {
                return $this->sendError('Transaction data not found.', [], 406);
            }

        } catch (\Exception $e) {
            \Log::error($e);
        }
    }

    public function userSearchForm(Request $request)
    {
        try {
            $data = $request->all();
            $user = \Auth::user();
            if ($user->role != 'customer') {
                $userform = view('pages/users/search')->render();
                return $this->sendResponse(['user_form' => $userform], 'User form.');
            }

        } catch (\Exception $e) {
            \Log::error($e);
        }
    }

    private function processTxnForm($user=[], $uid=0, $txnid=0)
    {
        try {
            if ($user->role != 'customer') {
                $mode = $txnid > 0 ? 'edit' : 'create';
                $txn = [];
                $services = Service::where('status', 1)->whereNull('deleted_at')->get();

                if ($mode == 'edit') {
                    if ((is_null($txnid) || $txnid == 0)) {
                        return $this->sendError('This request is not allowed.', [], 406);
                    } else {
                        $txn = Transaction::where('id', $txnid)->first();
                    }

                }

                $customer = [];
                $qrcode = null;
                $points = 0;
                if ($uid > 0) {
                    $customer = User::where('id', $uid)->with('profile')->first();
                    $points = UserHelper::getTotalPoints($uid);
                    $qrcode = UserHelper::generateQrCode($customer->profile->qrstring, 120);
                }

                // \Log::info('CUSTOMER', (array) $customer);
                $userform = view('pages/transactions/txn-create-edit', [
                    'mode'  => $mode,
                    'customer'  => $customer,
                    'fullname' => trim($customer->profile->firstname .' '.$customer->profile->lastname),
                    'txn' => $txn,
                    'qrcode'=> $qrcode,
                    'points' => $points,
                    'services' => $services
                ])->render();

                return $this->sendResponse(['user_form' => $userform], 'User form.');

            } else {
                return $this->sendError('This request is not allowed.', [], 406);
            }

        } catch (\Exception $e) {
            \Log::error($e);
        }
    }

    public function txnForm(Request $request, $uid=0, $txnid=0)
    {
        try {
            $data = $request->all();
            $user = \Auth::user();
            return $this->processTxnForm($user, $uid, $txnid);

        } catch (\Exception $e) {
            \Log::error($e);
        }
    }

    public function txnFormByScanner(Request $request, $encrypted='', $txnid=0)
    {
        try {
            $data = $request->all();
            $user = \Auth::user();
            $decrypted = (new Utilities)->encdecData('de', $encrypted);
            $uid = $decrypted['user_id'];
            return $this->processTxnForm($user, $uid, $txnid);

        } catch (\Exception $e) {
            \Log::error($e);
        }
    }

    public function mobileVerify(Request $request)
    {
        try {
            $data = $request->all();
            $user = \Auth::user();
            $validator = Validator::make($data, [
                'mobile' => 'required',
            ]);

            if ($validator->fails()) {
                return $this->sendError('Check these fields first before completing this request.', $validator->errors()->first(), 406);
            }

            $profile = UserProfile::where('mobile', $data['mobile'])->first();
            if (is_null($profile)) {
                return $this->sendError('User not found based on the provided mobile number.', [], 406);
            }
            $user = \Auth::user();
            $uid = $profile->user_id;
            return $this->processTxnForm($user, $uid, 0);


        } catch (\Exception $e) {
            \Log::error($e);
        }
    }


    public function userQRForm(Request $request)
    {
        try {
            $preview = view('pages/transactions/qrpreview')->render();
            return $this->sendResponse(['user_form' => $preview], 'User form.');

        } catch (\Exception $e) {
            \Log::error($e);
        }
    }

    public function mobileForm(Request $request)
    {
        try {
            $mobile = view('pages/transactions/mobile')->render();
            return $this->sendResponse(['user_form' => $mobile], 'User form.');

        } catch (\Exception $e) {
            \Log::error($e);
        }
    }

}
