<?php

namespace App\Models;

use Illuminate\Contracts\Auth\MustVerifyEmail;
use Illuminate\Database\Eloquent\Factories\HasFactory;
use Illuminate\Foundation\Auth\User as Authenticatable;
use Illuminate\Notifications\Notifiable;
use Laravel\Sanctum\HasApiTokens;
use Illuminate\Database\Eloquent\SoftDeletes;

class User extends Authenticatable
{
    use HasFactory, Notifiable, HasApiTokens, SoftDeletes;
    protected $cacheFor = 180; // 3 minutes
    protected $dates = ['deleted_at'];
    public $timestamps = true;

    /**
     * The attributes that are mass assignable.
     *
     * @var array<int, string>
     */
    protected $fillable = [
        'name', 'email', 'password', 'username', 'role', 'api_key', 'google_id'
    ];

    /**
     * The attributes that should be hidden for serialization.
     *
     * @var array<int, string>
     */
    protected $hidden = [
        'password', 'remember_token', 'api_key', 'facebook_id', 'google_id', 'status', 'email_verified_at', 'created_at', 'updated_at', 'deleted_at',
    ];

    /**
     * The attributes that should be cast.
     *
     * @var array<string, string>
     */
    protected $casts = [
        'email_verified_at' => 'datetime',
    ];

    public function profile()
    {
        return $this->hasOne('App\Models\UserProfile');
    }

    public function followers()
    {
        return $this->hasMany('App\Models\Follower', 'follower_id');
    }

    public function followersCount()
    {
        return $this->followers()->count();
    }

    public function transactions()
    {
        return $this->hasMany('App\Models\Transaction', 'user_id');
    }

    public function transactionsCount()
    {
        return $this->transactions()->count();
    }

    public function followings()
    {
        return $this->hasMany('App\Models\Following', 'following_id');
    }

    public function followingsCount()
    {
        return $this->followings()->count();
    }

    public function post()
    {
        return $this->belongsTo('App\Models\Post', 'user_id', 'id');
    }

    public function posts()
    {
        return $this->hasMany('App\Models\Post')->orderBy('created_at', 'DESC')->limit(10);
    }

    public function vlogs()
    {
        return $this->hasMany('App\Models\UserVlog')->orderBy('created_at', 'DESC')->limit(10);
    }

    public function setPasswordAttribute($value)
    {
        $this->attributes['password'] = bcrypt($value);
    }

    public function getUserForExplore($ids=[])
    {
        $flat_ids    = implode(',', array_map('intval', $ids));
        return count($ids) > 0 ? \DB::select("
            SELECT
                u.id as id,
                u.id as user_id,
                up.display_name,
                up.at_name,
                up.bio,
                up.avatar,
                up.cover,
                true as followed,
                0 as followers,
                0 as following
            FROM
                users as u
                LEFT JOIN user_profiles as up ON up.user_id = u.id
            WHERE
                u.id IN ({$flat_ids})
        ") : [] ;
    }

    public function getUserForSearchExplore($keyword='', $skip=0, $limit=20, $user_id=0)
    {
        $keyword = strtolower($keyword);
        $query = "
            SELECT
                u.id as id,
                u.id as user_id,
                u.username,
                u.email,
                up.display_name,
                up.at_name,
                up.bio,
                up.avatar,
                up.cover,
                true as followed,
                0 as followers,
                0 as following
            FROM
                users as u
                LEFT JOIN user_profiles as up ON up.user_id = u.id
            WHERE
                u.username LIKE '%$keyword%' OR
                u.email LIKE '%$keyword%' OR
                up.display_name LIKE '%$keyword%' OR
                up.at_name LIKE '%$keyword%' OR
                up.bio LIKE '%$keyword%' AND
                (
                    SELECT
                        COUNT(bl.id)
                    FROM
                        blocks as bl
                    WHERE
                        u.id = bl.blocked_user AND bl.blocked_by = {$user_id} AND ISNULL(bl.deleted_at)
                    LIMIT 1
                ) = 0
        ";

        if (!is_null($limit)) {
            $query .= "LIMIT {$skip}, {$limit}";
        }
        return !empty($keyword) ? \DB::select($query) : [];

    }

    protected function userDataQuery($whereClause='')
    {
        return "
            SELECT
                u.id,
                u.email,
                u.last_login,
                u.verified_at,
                CONCAT(up.firstname, ' ', up.lastname) as name,
                FORMAT(
                    IFNULL(
                        (SELECT sum(t.points) FROM transactions as t WHERE t.user_id = u.id AND t.claimed < 1 AND ISNULL(t.deleted_at) GROUP BY t.user_id),
                        0
                    )
                , 0) as points,
                u.created_at
            FROM
                users as u
                LEFT JOIN user_profiles as up ON up.user_id = u.id
            WHERE
                {$whereClause}
        ";
    }

    public function getCashiers($qp=[])
    {
        $column_orderable   = ['email', 'name', 'last_login', 'verified_at'];
        $qp['column_text']  = $qp['column'] == '' ? $column_orderable[0] : $column_orderable[$qp['column']];
        $qp['take']         = $qp['get_total'] == true || $qp['keyword'] != '' ? 5000000 : $qp['take'];
        $whereClause        = "u.role = 'cashier' AND ISNULL(u.deleted_at)";

        //IF KEYWORD IS NOT EMPTY
        if (!empty($qp['keyword'])) {
            $keyword = $qp['keyword'];
            $whereClause .= " AND (
                u.email LIKE '%$keyword%' OR up.firstname LIKE '%$keyword%' OR up.lastname LIKE '%$keyword%'
            )";
        }

        $query = $this->userDataQuery($whereClause);
        $query .= "
            ORDER BY {$qp['column_text']} {$qp['order_type']}
            LIMIT {$qp['skip']}, {$qp['take']}
        ";

        return \DB::select($query);

    }

    public function getCustomers($qp=[])
    {
        $column_orderable   = ['email', 'name', 'points', 'verified_at'];
        $qp['column_text']  = $qp['column'] == '' ? $column_orderable[0] : $column_orderable[$qp['column']];
        $qp['take']         = $qp['get_total'] == true || $qp['keyword'] != '' ? 5000000 : $qp['take'];
        $whereClause        = "u.role = 'customer' AND ISNULL(u.deleted_at)";

        //IF KEYWORD IS NOT EMPTY
        if (!empty($qp['keyword'])) {
            $keyword = $qp['keyword'];
            $whereClause .= " AND (
                u.email LIKE '%$keyword%' OR up.firstname LIKE '%$keyword%' OR up.lastname LIKE '%$keyword%'
            )";
        }

        $query = $this->userDataQuery($whereClause);
        $query .= "
            ORDER BY {$qp['column_text']} {$qp['order_type']}
            LIMIT {$qp['skip']}, {$qp['take']}
        ";

        return \DB::select($query);

    }

}
