<?php

namespace Tests\Feature\SetupAuth;

use Faker\Factory;
use Faker\Generator;
use Illuminate\Foundation\Testing\RefreshDatabase;
use Illuminate\Foundation\Testing\WithFaker;
use Tests\TestCase;

class SecurityTest extends TestCase
{
    private Generator $faker;
    /**
     * Test for signup feature.
     */
    public function test_signup_failed_to_create_user(): void
    {
        try {
            $response = $this->json('POST', '/api/signup', []);
            $response
                ->assertStatus(406)
                ->assertJson([
                    'success' => false,
                    'message' => "Check these fields first before completing Sign up process."
                ]);
        } catch (\Exception $e) {
            \Log::error($e);
        }
    }

    public function test_signup_where_password_does_not_match(): void
    {
        try {
            $response = $this->json('POST', '/api/signup', [
                'email' => env('TEST_USEREMAIL'),
                'password' => 'demopass',
                'password_confirmation' => 'demopass567',
            ]);

            $response
                ->assertStatus(406)
                ->assertJson([
                    'success' => false,
                    'message' => "Check these fields first before completing Sign up process."
                ]);
        } catch (\Exception $e) {
            \Log::error($e);
        }
    }

    public function test_signup_successfully_created_user(): void
    {
        try {
            $faker = Factory::create();
            $response = $this->json('POST', '/api/signup', [
                'email' => env('TEST_USEREMAIL'),
                'firstname' => 'Super',
                'lastname' => 'Admin',
                'password' => 'demopass567',
                'password_confirmation' => 'demopass567',
            ]);
            $response
                ->assertStatus(200)
                ->assertJson([
                    'success' => true,
                    'data' => [],
                    'message' => 'One last thing. We sent a code to your email. Use that to verify your account to complete the sign up process. Thanks.'
                    // 'message' => 'Congratulations! You can now login.'
                ]);
        } catch (\Exception $e) {
            \Log::error($e);
        }
    }

    public function test_signup_not_accepting_duplicate_email(): void
    {
        try {
            $faker = Factory::create();
            $response = $this->json('POST', '/api/signup', [
                'email' => env('TEST_USEREMAIL'),
                'firstname' => 'Super',
                'last_name' => 'Admin',
                'password' => 'demopass567',
                'password_confirmation' => 'demopass567',
            ]);
            $response
                ->assertStatus(406)
                ->assertJson([
                    'success' => false,
                    'message' => "Check these fields first before completing Sign up process."
                ]);
        } catch (\Exception $e) {
            \Log::error($e);
        }
    }

    public function test_login_authentication_is_true(): void
    {
        try {
            $response = $this->post('/api/login', [
                'email' => env('TEST_USEREMAIL'),
                'password' => 'demopass567',
            ]);
            $this->assertAuthenticated();
        } catch (\Exception $e) {
            \Log::error($e);
        }
    }

    public function test_login_authentication_is_false(): void
    {
        try {
            $response = $this->json('POST', '/api/login', [
                'email' => env('TEST_USEREMAIL'),
                'password' => 'demopass'
            ]);
            $response
                ->assertStatus(406)
                ->assertJson([
                    'success' => false,
                    'message' => "Login failed. Email / Password is invalid"
                ]);

        } catch (\Exception $e) {
            \Log::error($e);
        }
    }

    public function test_verify_code_is_true(): void
    {
        try {
            $user = \App\Models\User::where('email', env('TEST_USEREMAIL'))->with('profile')->first();
            $profile = $user->profile;
            $response = $this->json('POST', '/api/verify_code', [
                'signup_code' => $profile->signup_code,
                'signup_serialize' => $profile->signup_serialize,
            ]);

            $response
                ->assertStatus(200)
                ->assertJson([
                    'success' => true,
                    // 'data' => [],
                    'message' => 'One last thing. We sent a code to your email. Use that to verify your account to complete the sign up process. Thanks.'
                ]);

        } catch (\Exception $e) {
            \Log::error($e);
        }
    }

    public function test_verify_code_is_parameters_empty(): void
    {
        try {
            $response = $this->json('POST', '/api/verify_code', [
                'signup_code' => '',
                'signup_serialize' => '',
            ]);
            $response
            ->assertStatus(406)
            ->assertJson([
                'success' => false,
                'message' => "Invalid code please call admin to resolve this issue. Thank you."
            ]);
        } catch (\Exception $e) {
            \Log::error($e);
        }
    }
    public function test_verify_code_is_false(): void
    {
        try {
            $response = $this->json('POST', '/api/verify_code', [
                'signup_code' => 912019,
                'signup_serialize' => '@0182823',
            ]);
            $response
            ->assertStatus(406)
            ->assertJson([
                'success' => false,
                'message' => "Invalid code please call admin to resolve this issue. Thank you."
            ]);
        } catch (\Exception $e) {
            \Log::error($e);
        }
    }

    public function test_password_reset_request_true(): void
    {
        try {
            $user = \App\Models\User::where('email', env('TEST_USEREMAIL'))->with('profile')->first();
            $response = $this->json('POST', '/api/request_reset', [
                'email' => $user->email
            ]);

            $response
                ->assertStatus(200)
                ->assertJson([
                    'success' => true,
                    // 'data' => [],
                    'message' => 'Password reset request successful.'
                ]);

        } catch (\Exception $e) {
            \Log::error($e);
        }
    }

    public function test_password_reset_request_false(): void
    {
        try {
            $user = \App\Models\User::where('email', env('TEST_USEREMAIL'))->with('profile')->first();
            $response = $this->json('POST', '/api/request_reset', [
                'email' => 'user_'.$user->email
            ]);

            $response
                ->assertStatus(406)
                ->assertJson([
                    'success' => false,
                    // 'data' => [],
                    'message' => 'Check these fields first before completing this request.'
                ]);

        } catch (\Exception $e) {
            \Log::error($e);
        }
    }

    public function test_verify_reset_true(): void
    {
        try {
            $passwordReset = \App\Models\PasswordReset::where('email', env('TEST_USEREMAIL'))->first();
            $response = $this->json('POST', '/api/verify_reset', [
                'email' => $passwordReset->email,
                'token' => $passwordReset->token
            ]);

            $response
                ->assertStatus(200)
                ->assertJson([
                    'success' => true,
                    // 'data' => [],
                    'message' => 'Reset request is valid.'
                ]);

        } catch (\Exception $e) {
            \Log::error($e);
        }
    }

    public function test_verify_reset_false(): void
    {
        try {
            $response = $this->json('POST', '/api/verify_reset', [
                'email' => 'user_email@mail.com',
                'token' => 'mocktoken123'
            ]);

            $response
                ->assertStatus(406)
                ->assertJson([
                    'success' => false,
                    // 'data' => [],
                    'message' => 'Check these fields first before completing this request.'
                ]);

        } catch (\Exception $e) {
            \Log::error($e);
        }
    }

    public function test_password_reset_true(): void
    {
        try {
            $response = $this->json('POST', '/api/password_reset', [
                'email' => env('TEST_USEREMAIL'),
                'password' => 'P@s5w0d4dm17',
                'password_confirmation' => 'P@s5w0d4dm17',
            ]);

            $response
                ->assertStatus(200)
                ->assertJson([
                    'success' => true,
                    // 'data' => [],
                    'message' => 'Your password is now updated.'
                ]);

        } catch (\Exception $e) {
            \Log::error($e);
        }
    }

    public function test_password_reset_false(): void
    {
        try {
            $passwordReset = \App\Models\PasswordReset::where('email', env('TEST_USEREMAIL'))->first();
            $response = $this->json('POST', '/api/password_reset', [
                'email' => env('TEST_USEREMAIL'),
                'password' => 'P@s5w0d4dm14',
                'password_confirmation' => 'P@s5w0d4dm17',
            ]);

            $response
                ->assertStatus(406)
                ->assertJson([
                    'success' => false,
                    // 'data' => [],
                    'message' => 'Check these fields first before completing this request.'
                ]);

        } catch (\Exception $e) {
            \Log::error($e);
        }
    }

    public function test_login_authentication_after_password_change_is_true(): void
    {
        try {
            $response = $this->post('/api/login', [
                'email' => env('TEST_USEREMAIL'),
                'password' => 'P@s5w0d4dm17',
            ]);
            $this->assertAuthenticated();
        } catch (\Exception $e) {
            \Log::error($e);
        }
    }

    public function test_login_authentication_after_password_change_is_false(): void
    {
        try {
            $response = $this->post('/api/login', [
                'email' => env('TEST_USEREMAIL'),
                'password' => 'demopass567',
            ]);
            $response
                ->assertStatus(406)
                ->assertJson([
                    'success' => false,
                    'message' => "Login failed. Email / Password is invalid"
                ]);
        } catch (\Exception $e) {
            \Log::error($e);
        }
    }


}
